#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#


import sys
import time
from typing import Optional

import backoff
from airbyte_cdk.logger import AirbyteLogger
from requests import codes, exceptions

from .exceptions import DefaultBackoffException, UserDefinedBackoffException

TRANSIENT_EXCEPTIONS = (DefaultBackoffException, exceptions.ConnectTimeout, exceptions.ReadTimeout, exceptions.ConnectionError)

# TODO inject singleton logger?
logger = AirbyteLogger()


def default_backoff_handler(max_tries: Optional[int], factor: float, **kwargs):
    def log_retry_attempt(details):
        _, exc, _ = sys.exc_info()
        if exc.response:
            logger.info(f"Status code: {exc.response.status_code}, Response Content: {exc.response.content}")
        logger.info(
            f"Caught retryable error '{str(exc)}' after {details['tries']} tries. Waiting {details['wait']} seconds then retrying..."
        )

    def should_give_up(exc):
        # If a non-rate-limiting related 4XX error makes it this far, it means it was unexpected and probably consistent, so we shouldn't back off
        give_up = exc.response is not None and exc.response.status_code != codes.too_many_requests and 400 <= exc.response.status_code < 500
        if give_up:
            logger.info(f"Giving up for returned HTTP status: {exc.response.status_code}")
        return give_up

    return backoff.on_exception(
        backoff.expo,
        TRANSIENT_EXCEPTIONS,
        jitter=None,
        on_backoff=log_retry_attempt,
        giveup=should_give_up,
        max_tries=max_tries,
        factor=factor,
        **kwargs,
    )


def user_defined_backoff_handler(max_tries: Optional[int], **kwargs):
    def sleep_on_ratelimit(details):
        _, exc, _ = sys.exc_info()
        if isinstance(exc, UserDefinedBackoffException):
            if exc.response:
                logger.info(f"Status code: {exc.response.status_code}, Response Content: {exc.response.content}")
            retry_after = exc.backoff
            logger.info(f"Retrying. Sleeping for {retry_after} seconds")
            time.sleep(retry_after + 1)  # extra second to cover any fractions of second

    def log_give_up(details):
        _, exc, _ = sys.exc_info()
        logger.error(f"Max retry limit reached. Request: {exc.request}, Response: {exc.response}")

    return backoff.on_exception(
        backoff.constant,
        UserDefinedBackoffException,
        interval=0,  # skip waiting, we'll wait in on_backoff handler
        on_backoff=sleep_on_ratelimit,
        on_giveup=log_give_up,
        jitter=None,
        max_tries=max_tries,
        **kwargs,
    )



#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, List, Mapping

from airbyte_cdk.sources import Source
from airbyte_cdk.utils.mapping_utils import all_key_pairs_dot_notation, get_value_by_dot_notation


def get_secrets(source: Source, config: Mapping[str, Any], logger: logging.Logger) -> List[Any]:
    """
    Get a list of secrets from the source config based on the source specification
    """
    flattened_key_values = all_key_pairs_dot_notation(source.spec(logger).connectionSpecification.get("properties", {}))
    secret_key_names = [
        ".".join(key.split(".")[:1]) for key, value in flattened_key_values.items() if value and key.endswith("airbyte_secret")
    ]
    return [str(get_value_by_dot_notation(config, key)) for key in secret_key_names if config.get(key)]



#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#

import logging
import logging.config
import sys
import traceback
from typing import List, Tuple

from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage
from deprecated import deprecated

TRACE_LEVEL_NUM = 5

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "airbyte": {"()": "airbyte_cdk.logger.AirbyteLogFormatter", "format": "%(message)s"},
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "formatter": "airbyte",
        },
    },
    "root": {
        "handlers": ["console"],
    },
}


def init_unhandled_exception_output_filtering(logger: logging.Logger) -> None:
    """
    Make sure unhandled exceptions are not printed to the console without passing through the Airbyte logger and having
    secrets removed.
    """

    def hook_fn(exception_type, exception_value, traceback_):
        # For developer ergonomics, we want to see the stack trace in the logs when we do a ctrl-c
        if issubclass(exception_type, KeyboardInterrupt):
            sys.__excepthook__(exception_type, exception_value, traceback_)
        else:
            logger.critical(exception_value, exc_info=exception_value)

    sys.excepthook = hook_fn


def init_logger(name: str = None):
    """Initial set up of logger"""
    logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")
    logger = logging.getLogger(name)
    logger.setLevel(TRACE_LEVEL_NUM)
    logging.config.dictConfig(LOGGING_CONFIG)
    init_unhandled_exception_output_filtering(logger)
    return logger


class AirbyteLogFormatter(logging.Formatter):
    """Output log records using AirbyteMessage"""

    _secrets: List[str] = []

    @classmethod
    def update_secrets(cls, secrets: List[str]):
        """Update the list of secrets to be replaced in the log message"""
        cls._secrets = secrets

    # Transforming Python log levels to Airbyte protocol log levels
    level_mapping = {
        logging.FATAL: "FATAL",
        logging.ERROR: "ERROR",
        logging.WARNING: "WARN",
        logging.INFO: "INFO",
        logging.DEBUG: "DEBUG",
        TRACE_LEVEL_NUM: "TRACE",
    }

    def format(self, record: logging.LogRecord) -> str:
        """Return a JSON representation of the log message"""
        message = super().format(record)
        airbyte_level = self.level_mapping.get(record.levelno, "INFO")
        for secret in AirbyteLogFormatter._secrets:
            message = message.replace(secret, "****")
        log_message = AirbyteMessage(type="LOG", log=AirbyteLogMessage(level=airbyte_level, message=message))
        return log_message.json(exclude_unset=True)


def log_by_prefix(msg: str, default_level: str) -> Tuple[int, str]:
    """Custom method, which takes log level from first word of message"""
    valid_log_types = ["FATAL", "ERROR", "WARN", "INFO", "DEBUG", "TRACE"]
    split_line = msg.split()
    first_word = next(iter(split_line), None)
    if first_word in valid_log_types:
        log_level = logging.getLevelName(first_word)
        rendered_message = " ".join(split_line[1:])
    else:
        log_level = logging.getLevelName(default_level)
        rendered_message = msg

    return log_level, rendered_message


@deprecated(version="0.1.47", reason="Use logging.getLogger('airbyte') instead")
class AirbyteLogger:
    def log(self, level, message):
        log_record = AirbyteLogMessage(level=level, message=message)
        log_message = AirbyteMessage(type="LOG", log=log_record)
        print(log_message.json(exclude_unset=True))

    def fatal(self, message):
        self.log("FATAL", message)

    def exception(self, message):
        message = f"{message}\n{traceback.format_exc()}"
        self.error(message)

    def error(self, message):
        self.log("ERROR", message)

    def warn(self, message):
        self.log("WARN", message)

    def info(self, message):
        self.log("INFO", message)

    def debug(self, message):
        self.log("DEBUG", message)

    def trace(self, message):
        self.log("TRACE", message)



#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#

import contextlib
import os
import re
from typing import Any, Callable, List, Optional, Type, Union
from uuid import uuid4

import sentry_sdk
from sentry_sdk.integrations.atexit import AtexitIntegration
from sentry_sdk.integrations.excepthook import ExcepthookIntegration
from sentry_sdk.integrations.logging import LoggingIntegration


class AirbyteSentry:
    """
    Class for working with sentry sdk. It provides methods to:
        - init sentry sdk based on env variable
        - add breadcrumbs and set context
        - work with transactions and transaction spans
        - set tag and capture message and capture exception
    Also it implements client side sensitive data scrubbing.
    """

    DSN_ENV_NAME = "SENTRY_DSN"
    SECRET_MASK = "***"
    # Maximum number of breadcrumbs to send on fail. Breadcrumbs is trail of
    # events that occured before the fail and being sent to server only
    # if handled or unhandled exception occured.
    MAX_BREADCRUMBS = 30
    # Event sending rate. could be from 0 (0%) to 1.0 (100 % events being sent
    # to sentry server)
    TRACES_SAMPLE_RATE = 1.0
    SECRET_REGEXP = [
        re.compile("(api_key=)[a-zA-Z0-9_]+"),
        re.compile("(access_token=)[a-zA-Z0-9_]+"),
        re.compile("(refresh_token=)[a-zA-Z0-9_]+"),
        re.compile("(token )[a-zA-Z0-9_]+"),
        re.compile("(Bearer )[a-zA-Z0-9_]+"),
    ]
    SENSITIVE_KEYS = ["Authorization", "client_secret", "access_token"]

    sentry_enabled = False
    source_tag = ""
    run_id = str(uuid4())
    secret_values: List[str] = []

    @classmethod
    def process_value(cls, key: str, value: str):
        """
        Process single value. Used by recursive replace_value method or
        standalone for single value.
        """
        for secret in cls.secret_values:
            value = value.replace(secret, cls.SECRET_MASK)
        if key in cls.SENSITIVE_KEYS:
            return cls.SECRET_MASK
        for regexp in cls.SECRET_REGEXP:
            value = regexp.sub(f"\\1{cls.SECRET_MASK}", value)
        return value

    @classmethod
    def replace_value(cls, key, value):
        """
        Recursively scan event and replace all sensitive data with SECRET_MASK.
        Perform inplace data replace i.e. its not creating new object.
        """
        if isinstance(value, dict):
            for k, v in value.items():
                value[k] = cls.replace_value(k, v)
        elif isinstance(value, list):
            for index, v in enumerate(value):
                value[index] = cls.replace_value(index, v)
        elif isinstance(value, str):
            return cls.process_value(key, value)
        return value

    @classmethod
    def filter_event(cls, event, hint):
        """
        Callback for before_send sentry hook.
        """
        if "message" in event:
            event["message"] = cls.process_value(None, event["message"])
        cls.replace_value(None, event.get("exception"))
        cls.replace_value(None, event.get("contexts"))
        return event

    @classmethod
    def filter_breadcrumb(cls, event, hint):
        """
        Callback for before_breadcrumb sentry hook.
        """
        cls.replace_value(None, event)
        return event

    @classmethod
    def init(
        cls,
        source_tag: str = None,
        transport: Optional[Union[Type[sentry_sdk.transport.Transport], Callable[[Any], None]]] = None,
        secret_values: List[str] = [],
    ):
        """
        Read sentry data source name (DSN) from env variable and initialize sentry cdk.
        Args:
            source_tag: str -  Source name to be used in "source" tag for events organazing.
            transport: Transport or Callable - transport object for transfering
            sentry event to remote server. Usually used for testing, by default
            HTTP transport used
            secret_values: List[str] - list of string that have to be filtered
            out before sending event to sentry server.

        """
        sentry_dsn = os.environ.get(cls.DSN_ENV_NAME)
        if sentry_dsn:
            cls.sentry_enabled = True
            cls.secret_values = secret_values
            sentry_sdk.init(
                sentry_dsn,
                max_breadcrumbs=cls.MAX_BREADCRUMBS,
                traces_sample_rate=cls.TRACES_SAMPLE_RATE,
                before_send=AirbyteSentry.filter_event,
                before_breadcrumb=AirbyteSentry.filter_breadcrumb,
                transport=transport,
                # Use only limited list of integration cause sentry may send
                # transaction events e.g. it could send httplib request with
                # url and authorization info over StdlibIntegration and it
                # would bypass before_send hook.
                integrations=[
                    ExcepthookIntegration(always_run=True),
                    AtexitIntegration(),
                    LoggingIntegration(),
                ],
                # Disable default integrations cause sentry does not allow to
                # filter transactions event that could transfer sensitive data
                default_integrations=False,
            )
            if source_tag:
                sentry_sdk.set_tag("source", source_tag)
                sentry_sdk.set_tag("run_id", cls.run_id)
                cls.source_tag = source_tag

    def if_enabled(f):
        def wrapper(cls, *args, **kvargs):
            if cls.sentry_enabled:
                return f(cls, *args, **kvargs)

        return wrapper

    def if_enabled_else(return_value):
        def if_enabled(f):
            def wrapper(cls, *args, **kvargs):
                if cls.sentry_enabled:
                    return f(cls, *args, **kvargs)
                else:
                    return return_value

            return wrapper

        return if_enabled

    # according to issue CDK: typing errors #9500, mypy raises error on this line
    # 'Argument 1 to "if_enabled" has incompatible type "Callable[[Type[AirbyteSentry], str, Any], Any]"; expected "AirbyteSentry"'
    # there are a few similar opened issues
    # https://github.com/python/mypy/issues/12110
    # https://github.com/python/mypy/issues/11619
    # ignored for now
    @classmethod  # type: ignore
    @if_enabled
    def set_tag(cls, tag_name: str, value: Any):
        """
        Set tag that is handy for events organazing and filtering by sentry UI.
        """
        sentry_sdk.set_tag(tag_name, value)

    # same ignored as for line 171
    @classmethod  # type: ignore
    @if_enabled
    def add_breadcrumb(cls, message, data=None):
        """
        Add sentry breadcrumb.
        """
        sentry_sdk.add_breadcrumb(message=message, data=data)

    # same ignored as for line 171
    @classmethod  # type: ignore
    @if_enabled
    def set_context(cls, name, data):
        # Global context being used by transaction event as well. Since we cant
        # filter senstitve data coming from transaction event using sentry
        # before_event hook, apply filter to context here.
        cls.replace_value(None, data)
        sentry_sdk.set_context(name, data)

    # same ignored as for line 171
    @classmethod  # type: ignore
    @if_enabled
    def capture_message(cls, message):
        """
        Send message event to sentry.
        """
        sentry_sdk.capture_message(message)

    # same ignored as for line 171
    @classmethod  # type: ignore
    @if_enabled
    def capture_exception(
        cls,
        error: Optional[BaseException] = None,
        scope: Optional[Any] = None,
        **scope_args,
    ):
        """
        Report handled execption to sentry.
        """
        sentry_sdk.capture_exception(error, scope=scope, **scope_args)

    # same ignored as for line 171
    @classmethod
    @if_enabled_else(contextlib.nullcontext())  # type: ignore
    def start_transaction(cls, op, name=None):
        """
        Return context manager for starting sentry transaction for performance monitoring.
        """
        return sentry_sdk.start_transaction(op=op, name=f"{cls.source_tag}.{name}")

    # same ignored as for line 171
    @classmethod
    @if_enabled_else(contextlib.nullcontext())  # type: ignore
    def start_transaction_span(cls, op, description=None):
        """
        Return context manager for starting sentry transaction span inside existing sentry transaction.
        """
        # Apply filter to description since we cannot use before_send sentry
        # hook for transaction event.
        description = cls.replace_value(None, description)
        return sentry_sdk.start_span(op=op, description=description)



#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#

from functools import reduce
from typing import Any, Iterable, List, Mapping, Optional, Tuple


def all_key_pairs_dot_notation(dict_obj: Mapping) -> Mapping[str, Any]:
    """
    Recursively iterate through a dictionary and return a dictionary of all key-value pairs in dot notation.
    keys are prefixed with the list of keys passed in as prefix.
    """

    def _all_key_pairs_dot_notation(_dict_obj: Mapping, prefix: List[str] = []) -> Iterable[Tuple[str, Any]]:
        for key, value in _dict_obj.items():
            if isinstance(value, dict):
                prefix.append(str(key))
                yield from _all_key_pairs_dot_notation(value, prefix)
                prefix.pop()
            else:
                prefix.append(str(key))
                yield ".".join(prefix), value
                prefix.pop()

    return {k: v for k, v in _all_key_pairs_dot_notation(dict_obj)}


def get_value_by_dot_notation(dict_obj: Mapping, key: str, default: Optional[Any] = ...) -> Any:
    """
    Return the value of a key in dot notation in a arbitrarily nested Mapping.
    dict_obj: Mapping
    key: str
    default: Any
    raises: KeyError if default is not provided and the key is not found
    ex.:
        dict_obj = {"nested": {"key": "value"}}
        get_value_by_dot_notation(dict_obj, "nested.key") == "value" -> True
    """

    return reduce(lambda d, key_name: d[key_name] if default is ... else d.get(key_name, default), key.split("."), dict_obj)
